/* pclta.c: Device driver for PCLTA LON-interface card */
/*
        Written 1997-1999 by Miran Miksic
        Copyright 1997-1999 Miran Miksic
        Copyright 1997-1999 Fastec GmbH

        The author may be reached as mmiksic@hni.uni-paderborn.de.
        Heinz Nixdorf Institut, Universitaet GH Paderborn
        Fuerstenallee 11, 33102 Paderborn, Germany

    Revision: $Revision: 2.2 $
    RCS: $Id: pclta.c,v 2.2 1999/01/06 14:23:58 miksic Exp $
*/

static char *version =
        "pclta: pclta driver v2.04a $Date: 1999/01/06 14:23:58 $\n"
        "Copyright 1997-1999 Miran Miksic (mmiksic@hni.uni-paderborn.de)\n";

/************/
/* includes */
/************/

#include <linux/config.h>
    /* include the configuration header file, to be certain that a module
     * is compiled together with the actual version of kernel
     * but, it seems not to work at the moment */

#if defined MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
    /* include module and version headers */

#include <linux/kernel.h>        /* for KERN_WARNING, ... */
#include <linux/malloc.h>        /* for kmalloc(), kfree(), ... */
#include <linux/mm.h>            /* for verify_area(), ... */
#include <linux/ioport.h>        /* for *_region() */
#include <asm/io.h>              /* for in*(), out*() */
#include <asm/irq.h>             /* for disable_irq(), enable_irq() */

#include "pclta.h"

/***********/
/* defines */
/***********/

#define MAX_DEV 6        /* maximum number of devices */

/*************/
/* variables */
/*************/

/* --------------------------- */
/* input parameters for insmod */
/* --------------------------- */

static int major = 40;
    /* major number of the device */

static int probe_io[MAX_DEV] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static int probe_irq[MAX_DEV] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
    /* base io addresses and interrupts for probing; device needs four
     * I/O addresses for every channel; the corresponding minor numbers
     * are derived from the order of declaration */

/* --------------- */
/* local variables */
/* --------------- */

static struct pclta_device *device_table[MAX_DEV] =
    { NULL, NULL, NULL, NULL, NULL, NULL };
    /* array of initiated devices */

static unsigned int irq_table[16] =
    { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
    /* an entry in the table corresponds to the minor device number to which
     * this interrupt was assigned */

static void timer_handler( unsigned long data );
static struct timer_list timer_list =
    { NULL, NULL, 0, 0, timer_handler };
    /* timer list to activate timer */

static volatile int timeout;
    /* variable to indicate if timeout occured */

/**************/
/* prototypes */
/**************/

/* ------------------- */
/* init/cleanup module */
/* ------------------- */

#if !defined MODULE
long int pclta_init( long mem_start, long mem_end );
#endif

int init_module( void );
static boolean probe_hw( unsigned int minor );
static void init_hw( unsigned int port, unsigned char irq,
    unsigned char clock );

#if defined MODULE
void cleanup_module( void );
static void cleanup_hw( struct pclta_device *dev );
#endif

/* -------------------- */
/* registered functions */
/* -------------------- */

static int read_pclta( struct inode *inode, struct file *file,
    char *buf, int cnt );
static int write_pclta( struct inode *inode, struct file *file,
    const char *buf, int cnt );
static int select_pclta( struct inode *inode, struct file *file,
    int sel_type, select_table *wait );
static int open_pclta( struct inode *inode, struct file *file );
static void close_pclta( struct inode *inode, struct file *file );

static struct file_operations pclta_fops =
{
    NULL,                /* lseek() */
    read_pclta,          /* read() */
    write_pclta,         /* write() */
    NULL,                /* readdir() */
    select_pclta,        /* select() */
    NULL,                /* ioctl() */
    NULL,                /* mmap() */
    open_pclta,          /* open() */
    close_pclta,         /* release() */
    NULL,                /* fsync() */
    NULL,                /* fasync() */
    NULL,                /* check_media_change() */
    NULL                 /* revalidate */
};

/* ---------------------------*/
/* timer dependent read/write */
/* ---------------------------*/

static boolean write_byte_timeout( unsigned int port, unsigned char byte,
    int sec100 );
    /* write single byte, fails if timer expires */

static boolean wait_reset( unsigned int port, int sec100 );

static boolean wait_uplink_buffer( unsigned int port, int sec100,
    unsigned char *buf );
    /* wait for card to provide an uplink buffer, read buffer */

static void setup_timeout( int sec100 );
    /* setup timer to expire in sec100 miliseconds */

/* ------------------------------- */
/* callbacks for interrupt handler */
/* ------------------------------- */

static void pclta_start_downlink( struct pclta_device *dev );
static void pclta_service( struct pclta_device *dev );

/* ------------------ */
/* interrupt handlers */
/* ------------------ */

static void pclta_interrupt( int irq, void *dev_id, struct pt_regs *regs );
static void timer_handler( unsigned long data );

/* ---------------- */
/* block read/write */
/* ---------------- */

static boolean uplink_packet( unsigned int port, unsigned char *buf );
static void downlink_packet( unsigned int port, unsigned char *buf );

/* ---------------------- */
/* inline byte read/write */
/* ---------------------- */

static inline unsigned char read_byte_wait( unsigned int port );
static inline void write_byte_wait( unsigned int port, unsigned char byte );

/* -------------------------- */
/* device structure functions */
/* -------------------------- */

static struct pclta_device *create_device( unsigned int io, unsigned int irq,
    unsigned char clock );
static void remove_device( struct pclta_device *dev );
static boolean open_device( struct pclta_device *dev );
static void close_device( struct pclta_device *dev );
static void init_buffer( struct pclta_buffer *buf );

/******************/
/* implementation */
/******************/

/* ------------------- */
/* init/cleanup module */
/* ------------------- */

#if !defined MODULE
long int
pclta_init( long mem_start, long mem_end )
{
    init_module();
    return mem_start;
}
#endif

int
init_module( void )
{
    unsigned int i, count=0;

    printk( KERN_WARNING "%s", version );
    for( i=0; i<MAX_DEV; i++ ) {
        if( probe_hw( i ) ) count++;
    }
    if( count ) {
        printk( KERN_WARNING "%d device(s) detected.\n", count );
        return 0;
    } else {
        printk( KERN_WARNING "No devices detected.\n" );
        return -ENXIO;
    }
}

static boolean
probe_hw( unsigned int minor )
{
    int erc, trans;
    unsigned char rx_buffer[256];
    unsigned int io=probe_io[minor], irq=probe_irq[minor];
    struct pclta_device *dev;

    if( !io || !irq ) return 0;

    printk( KERN_WARNING "probing minor=%d, I/O=%#x-%#x, irq=%d ... ",
        minor, io, io+3, irq );
    if( 0 > check_region( io, 4 ) ) {
        printk( "error(%d).\n", PCLTA_IOPERM );
        return false;
    }

    init_hw( io, 0x00, CLK_10 ); /* init hardware */

    if( ! wait_reset( io, 500 ) ) {
        printk( "error(%d).\n", PCLTA_LRESET );
        return false;
    } /* waiting to leave reset state */

    setup_timeout( 200 );
    while( !timeout );
    /* Why wait? Because, when PCLTA gets out of RESET state, it sets
     * status register BEFORE preparing uplink reset. So, I read the
     * packet of length 0 and get an error. Stupid! */

    if( ! wait_uplink_buffer( io, 500, rx_buffer ) ) {
        printk( "error(%d).\n", PCLTA_RESP );
        return false;
    } /* wait for uplink reset */

    if( rx_buffer[0] != 1 || rx_buffer[1] != niRESET ) {
        printk( "error(%d).\n", PCLTA_RESP );
        return false;
    } /* check uplink reset */

    write_byte_wait( io, CMD_XFER );
    write_byte_wait( io, 2 );
    write_byte_wait( io, niIRQENA );
    write_byte_wait( io, IRQPLRTY | RESET_E | DLREADY_E | DLPBA_E | DLBA_E | ULREADY_E );
        /* set up internal interrupt register */

    write_byte_wait( io, CMD_XFER );
    write_byte_wait( io, 1 );
    write_byte_wait( io, niSSTATUS );
        /* write a command to get transciever status */

    if( ! wait_uplink_buffer( io, 500, rx_buffer ) ) {
        printk( "error(%d).\n", PCLTA_STAT );
        return false;
    }

    trans = rx_buffer[2]>>3;
    if( ! *transceiver[trans] ) {
        printk( "error(%d).\n", PCLTA_TSCV );
        return false;
    } /* transciever not supported */

    if( 0 > register_chrdev( major, "pclta", &pclta_fops ) ) {
        printk( "error(%d).\n", PCLTA_MAJOR );
        return false;
    }
    if( 0 > ( erc = request_irq( irq, pclta_interrupt, 0, "pclta", NULL ) ) ) {
        printk( "error(%d).\n", PCLTA_IRQPERM );
        return false;
    }
    dev = create_device( io, irq, clock[rx_buffer[2]>>3] );
    if( dev == NULL ) {
        printk( "error(%d).\n", PCLTA_CDEV );
        return false;
    }
    device_table[minor] = dev; /* device registered */
    irq_table[irq] = minor; /* now we know where to search on interrupt */
    request_region( io, 4, "pclta" );
    printk( "(%s) done.\n", transceiver[trans] );
    return true;
}

static void
init_hw( unsigned int port, unsigned char irq, unsigned char clock )
{
    outb_p( irq, _selirq_reg(port) ); /* interrupt off */
    outb_p( clock, _clkrst_reg(port) );
    outb_p( _RESET | clock, _clkrst_reg(port) ); /* reset device */
    if( irq ) outb_p( ~CLRIRQ, _clrirq_reg(port) ); /* ack. IRQ line */
}

#if defined MODULE

void
cleanup_module( void )
{
    unsigned int i;

    for( i=0; i<MAX_DEV; i++ ) cleanup_hw( device_table[i] );
    unregister_chrdev( major, "pclta" );
}

static void
cleanup_hw( struct pclta_device *dev )
{
    if( dev==NULL ) return; /* quick return */

    /* write_byte_wait( dev->io, CMD_XFER );
     * write_byte_wait( dev->io, 2 );
     * write_byte_wait( dev->io, niIRQENA );
     * write_byte_wait( dev->io, IRQPLRTY ); */
        /* reset internal interrupt register */
        /* commented out because of the NT driver */

    release_region( dev->io, 4 );
    free_irq( dev->irq, NULL );
    remove_device( dev ); /* free memory */
}

#endif

/* -------------------- */
/* registered functions */
/* -------------------- */

static int
read_pclta( struct inode *inode, struct file *file, char *buf, int cnt )
{
    struct pclta_device *dev = device_table[MINOR(inode->i_rdev)];
    struct xfer_packet *xfer_buf;
    
/* verification */
    if( dev == NULL ) return -ENXIO;
    if( 0>verify_area( VERIFY_WRITE, buf, cnt ) ) return -EFAULT;

/* wait for uplink packet */
    xfer_buf = dev->rx->tail;
    if( !xfer_buf->busy && (file->f_flags & O_NONBLOCK) ) return -EWOULDBLOCK;
    while( !xfer_buf->busy ) {
        interruptible_sleep_on( &dev->wait_queue );
        if( current->signal & ~current->blocked )
            return -EINTR;
    }
    if( cnt <= xfer_buf->xfer_length ) return -EINVAL;

/* copy data from buffer to user space */
    memcpy_tofs( buf+2, xfer_buf->xfer_data, xfer_buf->xfer_length-1 );
    put_user( xfer_buf->xfer_cmd, buf );
    put_user( xfer_buf->xfer_length-1, buf+1 );

/* update buffer data and restart uplink if backed off */
    xfer_buf->busy = false;
    dev->rx->tail = xfer_buf->next;
    if( dev->restart_uplink ) {
        disable_irq( dev->irq );
        dev->restart_uplink = false;
        if( dev->state == IdleState )
            pclta_service( dev );
        enable_irq( dev->irq );
    } /* restart uplink, but only if device is in the IdleState. We don't
       * want to mess up waiting for an CMD_XFER interrupt, do we? */
    return xfer_buf->xfer_length+1;
}

static int
write_pclta( struct inode *inode, struct file *file, const char *buf, int cnt )
{
    struct pclta_device *dev = device_table[MINOR(inode->i_rdev)];
    struct xfer_packet *xfer_buf;
    unsigned int len;
    
/* verification */
    if( dev == NULL ) return -ENXIO;
    if( 0 > verify_area( VERIFY_READ, buf, cnt ) ) return -EFAULT;
    if( cnt < 2 ) return -EINVAL;

/* wait for free buffer entry */
    xfer_buf = dev->tx->head;
    if( xfer_buf->busy && (file->f_flags & O_NONBLOCK) ) return -EWOULDBLOCK;
    while( xfer_buf->busy ) {
        interruptible_sleep_on( &dev->wait_queue );
        if( current->signal & ~current->blocked )
            return -EINTR;
    }

/* copy from user space to buffer */
    xfer_buf->xfer_cmd = get_user( buf++ );
    len = xfer_buf->xfer_length = get_user( buf++ ) + 1;
    if( cnt <= len ) return -EINVAL;
    memcpy_fromfs( xfer_buf->xfer_data, buf, len-1 );

/* update output buffers and start downlink transfer */
    dev->tx->head = xfer_buf->next;
    xfer_buf->busy = true;
    if( dev->state == IdleState ) {
        disable_irq( dev->irq );
        pclta_start_downlink( dev );
        enable_irq( dev->irq );
    }
    return len+1;
}

static int
select_pclta( struct inode *inode, struct file *file, int sel_type,
    select_table *wait )
{
    struct pclta_device *dev = device_table[MINOR(inode->i_rdev)];

    if( dev == NULL ) return -ENXIO;
    switch( sel_type ) {
    case SEL_IN:
       if( dev->rx->tail->busy ) return 1;
       break;
    case SEL_OUT:
       if( ! dev->tx->head->busy ) return 1;
       break;
    case SEL_EX:
       return -EPERM;
    }
    select_wait( &dev->wait_queue, wait );
    return 0;
}

static int
open_pclta( struct inode *inode, struct file *file )
{
    struct pclta_device *dev = device_table[MINOR(inode->i_rdev)];

    if( dev == NULL ) return -ENXIO; /* no such device */
    if( dev->state != NotOpened ) return -EBUSY; /* device busy */
    if( !open_device( dev ) ) return -ENOMEM; /* out of memory */
    init_hw( dev->io, irq_mask[dev->irq], dev->clock );
    dev->state = IdleState;
    MOD_INC_USE_COUNT;
    return 0;
}

static void
close_pclta( struct inode *inode, struct file *file )
{
    struct pclta_device *dev = device_table[MINOR(inode->i_rdev)];

    if( dev == NULL ) return; /* no such device */
    if( dev->state == NotOpened ) return; /* device already closed */
    if( file->f_count>1 ) return; /* not the last process */
    dev->state = NotOpened;
    init_hw( dev->io, 0x00, dev->clock );
    close_device( dev );
    MOD_DEC_USE_COUNT;
}

/* ---------------------------*/
/* timer dependent read/write */
/* ---------------------------*/

static boolean
write_byte_timeout( unsigned int port, unsigned char byte, int sec100 )
{
    setup_timeout( sec100 );
    while( ( inb_p( _status_reg(port) ) & _READY ) && ( ! timeout ) );
    if( ! timeout ) outb_p( byte, _data_reg(port) );
    del_timer( &timer_list );
    return timeout ? false : true;
}

static boolean
wait_reset( unsigned int port, int sec100 )
{
    setup_timeout( sec100 );
    while( ( inb_p(_status_reg(port)) & RESET ) && !timeout );
    del_timer( &timer_list );
    if( timeout ) return false;
    return true;
}

static boolean
wait_uplink_buffer( unsigned int port, int sec100, unsigned char *buf )
{
    setup_timeout( sec100 );
    while( ! (inb_p(_status_reg(port)) & ULREADY) && !timeout );
    del_timer( &timer_list );
    if( timeout ) return false;
    uplink_packet( port, buf );
    return true;
}

static void
setup_timeout( int sec100 )
{
    timer_list.data = 0;
    init_timer( &timer_list );
    timer_list.expires = jiffies + sec100 * (HZ/100);
    add_timer( &timer_list );
    timeout = 0;
}

/* ------------------------------- */
/* callbacks for interrupt handler */
/* ------------------------------- */

/* if( state==IdleState AND downlink_packet_pending )
 * |   if( downlink_buffer_avaiable AND interface_not_reset )
 * |   |   start_downlink_transfer;
 * |   |   state = XferUnderway;
 * |   endif
 * endif */

static void
pclta_start_downlink( struct pclta_device *dev )
{
    struct xfer_packet *xfer_buf;
    unsigned char ni_cmd, buff_p, status;

    if( dev->state == IdleState ) {
        xfer_buf = dev->tx->tail;
        if( xfer_buf->busy ) {
            ni_cmd = xfer_buf->xfer_cmd;
            if( ni_cmd == niSERVICE ) {
                buff_p = DLPBA;
                ni_cmd = niCOMM;
            } else {
                buff_p = ni_cmd & niPRIORITY ? DLPBA : DLBA;
                ni_cmd &= NI_Q_CMD;
            }
            status = inb_p( status_reg(dev) );
            if( ni_cmd == niCOMM || ni_cmd == niNETMGMT ) {
                if( status & buff_p )
                    dev->state = XferUnderway;
            } else
                dev->state = XferUnderway;
            if( dev->state == XferUnderway ) {
                if( status & RESET )
                    dev->state = IdleState;
                else {
                    PCLTA_DEBUG( printk( "xfer_start, " ); )
                    write_byte_wait( dev->io, CMD_XFER );
                }
            }
        }
    }
}

/* read_device_state;
 * if( device_resetting )
 * |   state = IdleState
 * |   return;
 * endif
 * if( state==XferUnderway )
 * |   complete_downlink_transfer;
 * |   state = IdleState;
 * |   read_device_state;
 * if( uplink_packet_pending )    --- notice: state==IdleState
 * |   if( uplink_buffer_available )
 * |   |   uplink_packet;
 * |   else
 * |   |   restart_uplink=true;
 * |   endif
 * endif
 * start_downlink; */

static void
pclta_service( struct pclta_device *dev )
{
    unsigned int io = dev->io;
    struct xfer_packet *xfer_buf;
    unsigned char status;

    status = inb_p( _status_reg(io) );
    PCLTA_DEBUG( printk( "status: %#x, ", status ); )
    if( status & RESET ) {
        dev->state = IdleState;
        return;
    }
    if( dev->state == XferUnderway ) {
        PCLTA_DEBUG( printk( "xfer_body, " ); )
        xfer_buf = dev->tx->tail;
        downlink_packet( io, & xfer_buf->xfer_length );
        xfer_buf->busy = false;
        dev->tx->tail = xfer_buf->next;
        dev->state = IdleState;
        status = inb_p( _status_reg(io) );
        PCLTA_DEBUG( printk( "status: %#x, ", status ); )
    }
    if( status & ULREADY ) {
        xfer_buf = dev->rx->head;
        if( !xfer_buf->busy ) {
            if( uplink_packet( io, & xfer_buf->xfer_length ) ) {
                xfer_buf->busy = true;
                dev->rx->head = xfer_buf->next;
            }
        } else
            dev->restart_uplink = true;
    }
    pclta_start_downlink( dev );
}

/* ------------------ */
/* interrupt handlers */
/* ------------------ */

static void
pclta_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
    struct pclta_device *dev = device_table[ irq_table[irq] ];

    if( dev==NULL ) return; /* no such device ??? */
    if( dev->state==NotOpened ) return; /* Ooops ... something got lost */
    outb_p( ~CLRIRQ, clrirq_reg(dev) ); /* acknowledge interrupt */
    PCLTA_DEBUG( printk( "IRQ#%d: ", irq ); )
    pclta_service( dev ); /* service it */
    wake_up_interruptible( &dev->wait_queue ); /* it's time ... */
    PCLTA_DEBUG( printk( "IRQ#%d.\n", irq ); )
}

static void
timer_handler( unsigned long data )
{
    timeout = -1;
    return;
}

/* ---------------- */
/* block read/write */
/* ---------------- */

static boolean
uplink_packet( unsigned int port, unsigned char *buf )
{
    unsigned char len;

    write_byte_wait( port, CMD_ACK );
    read_byte_wait( port ); /* dummy */
    *buf++ = len = read_byte_wait( port );
    PCLTA_DEBUG( printk( "len %d, data ", len ); )
    if( len == 0 ) return false; /* drop it */
    while( len-- ) {
        *buf++ = read_byte_wait( port );
        PCLTA_DEBUG( printk( "[%2x] ", *(buf-1) ); )
    }
    return true;
}

static void
downlink_packet( unsigned int port, unsigned char *buf )
{
    unsigned char len = *buf;

    write_byte_wait( port, len );
    while( len-- ) write_byte_wait( port, *(++buf) );
}

/* ---------------------- */
/* inline byte read/write */
/* ---------------------- */

static inline unsigned char
read_byte_wait( unsigned int port )
{
    while( inb_p( _status_reg(port) ) & _READY );
    return inb_p( _data_reg(port) );
}

static inline void
write_byte_wait( unsigned int port, unsigned char byte )
{
    while( inb_p( _status_reg(port) ) & _READY );
    outb_p( byte, _data_reg(port) );
}

/* -------------------------- */
/* device structure functions */
/* -------------------------- */

static struct pclta_device *
create_device( unsigned int io, unsigned int irq, unsigned char clock )
{
    struct pclta_device *dev;

    dev = kmalloc( sizeof(struct pclta_device), GFP_KERNEL );
    if( dev == NULL ) return NULL;
    dev->io = io;
    dev->irq = irq;
    dev->state = NotOpened;
    dev->rx = dev->tx = NULL;
    dev->clock = clock;
    return dev;
}

static void
remove_device( struct pclta_device *dev )
{
    if( dev == NULL ) return;
    close_device( dev );
    kfree( dev );
}

static boolean
open_device( struct pclta_device *dev )
{
    dev->rx = kmalloc( sizeof(struct pclta_buffer), GFP_KERNEL );
    if( dev->rx == NULL ) return false;
    dev->tx = kmalloc( sizeof(struct pclta_buffer), GFP_KERNEL );
    if( dev->tx == NULL ) { kfree( dev->rx ); return false; }
    init_buffer( dev->rx );
    init_buffer( dev->tx );
    dev->restart_uplink = 0;
    dev->wait_queue = NULL;
    return true;
}

static void
close_device( struct pclta_device *dev )
{
    if( dev->state == NotOpened ) return;
    kfree( dev->rx );
    kfree( dev->tx );
    dev->tx = dev->rx = NULL;
    wake_up_interruptible( &dev->wait_queue );
}

static void
init_buffer( struct pclta_buffer *buf )
{
    unsigned int i;
    for( i=0; i<BUF_PACKETS; i++ ) {
        buf->data[i].next = & buf->data[(i+1)%BUF_PACKETS];
        buf->data[i].busy = false;
    }
    buf->head = buf->tail = buf->data;
}
/*
 * Local variables:
 *  compile-command: "gcc -Wall -O6 -N -D__KERNEL__ -DMODULE -DLINUX -o pclta.o -c pclta.c"
 *  tab-width: 8
 *  c-indent-level: 4
 * End:
 */
